##############################################################################
#
# Copyright (c) 2010 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

try:
    from zc.buildout.buildout import Buildout
except ImportError:
    raise ImportError("zc.buildout is not installed! \n"
          "If you're in a buildout environment,\n"
          "enable the buildout extra requirement like this:\n"
          "eggs = z3c.checkversions [buildout]")

from z3c.checkversions import base


class Checker(base.Checker):
    """checker class for a buildout
    """
    def __init__(self, *args, **kw):
        self.filename = kw.pop('filename')
        super(Checker, self).__init__(*args, **kw)

    def get_versions(self):
        print("# Checking buildout file %s" % self.filename)

        buildout = Buildout(self.filename, '')

        # set the index URL from the buildout if not already provided
        buildout_index = buildout['buildout'].get('index')
        if not self.__custom_url:
            self._set_index_url(buildout_index)

        return buildout['versions']





##############################################################################
#
# Copyright (c) 2010 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from pkg_resources import parse_version, Requirement
from setuptools import package_index


def _final_version(parsed_version):
    """Function copied from zc.buildout.easy_install._final_version
    """
    return not parsed_version.is_prerelease


class Checker(object):
    """Base class for version checkers

    attributes:

    index_url: url of an alternative package index
    verbose: display every version, not only new ones,
             and display the previous version as a comment
    blacklist: filename of the blacklist
    incremental: suggest only one package upgrade
    """
    __custom_url = False
    def __init__(self,
                 index_url=None,
                 verbose=False,
                 blacklist=None,
                 incremental=False):
        self.verbose = verbose
        self.incremental = incremental
        if blacklist:
            # create a set of tuples with bad versions
            with open(blacklist) as b:
                self.blacklist = set([
                    tuple(map(lambda x: x.strip(), line.split('=')))
                    for line in b.readlines()
                    if '=' in line
                ])
        else:
            self.blacklist = set()
        self.pi = package_index.PackageIndex(search_path=())
        self._set_index_url(index_url)
        if index_url is not None:
            self.__custom_url = True

    def _set_index_url(self, url):
        """set the index URL
        """
        if url is not None:
            self.pi.index_url = url
        if not self.pi.index_url.endswith('/'):
            self.pi.index_url += '/'

    def check(self, level=0):
        """Search new versions in a version list
        versions must be a dict {'name': 'version'}

        The new version is limited to the given level:
        Example with version x.y.z
        level = 0: checks new version x
        level = 1: checks new version y
        level = 2: checks new version z

        By default, the highest version is found.
        """
        versions = self.get_versions()

        for name, version in sorted(versions.items()):
            if self.incremental == 'stop':
                # skip subsequent scans
                print("%s=%s" % (name, version))
                continue
            parsed_version = parse_version(version)
            req = Requirement.parse(name)
            self.pi.find_packages(req)
            new_dist = None
            # loop all index versions until we find the 1st newer version
            # that keeps the major versions (below level)
            # and is a final version
            # and is not in the blacklist
            for dist in self.pi[req.key]:
                if self.incremental == 'stop':
                    continue
                if (dist.project_name, dist.version) in self.blacklist:
                    continue
                if _final_version(parsed_version) and not _final_version(dist.parsed_version):
                    #only skip non-final releases if the current release is a final one
                    continue
                # trunk the version tuple to the first `level` elements
                trunked_current = parsed_version.base_version.split('.')[:level]
                trunked_candidate = dist.parsed_version.base_version.split('.')[:level]
                while len(trunked_candidate) < level:
                    trunked_candidate.append('00000000')
                while len(trunked_current) < level:
                    trunked_current.append('00000000')
                # ok now we can compare: -> skip if we're still higher.
                if trunked_candidate > trunked_current:
                    continue
                new_dist = dist
                break

            if new_dist and new_dist.parsed_version > parsed_version:
                if self.incremental == True:
                    self.incremental = 'stop'
                if self.verbose:
                    print("%s=%s # was: %s" % (name, new_dist.version, version))
                else:
                    print("%s=%s" % (name, new_dist.version))
            elif self.verbose:
                print("%s=%s" % (name, version))


    def get_versions(self):
        """Get a dict {'name': 'version', ...} with package versions to check.
        This should be implemented by derived classes
        """
        raise NotImplementedError




##############################################################################
#
# Copyright (c) 2010 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from doctest import DocFileSuite, ELLIPSIS, NORMALIZE_WHITESPACE
import distutils.log
import os
import tempfile

def write_temp_file(content):
    fd, path = tempfile.mkstemp(prefix='test-z3c.checkversions-')
    f = os.fdopen(fd, 'w')
    f.write(content)
    f.close()
    return path

def setUp(test):
    test._old_log_level = distutils.log.set_threshold(distutils.log.ERROR)

def tearDown(test):
    distutils.log.set_threshold(test._old_log_level)

def test_suite():
    optionflags = ELLIPSIS|NORMALIZE_WHITESPACE
    suite = DocFileSuite('README.txt', 'buildout.txt', 'installed.txt',
                         setUp=setUp, tearDown=tearDown,
                         optionflags=optionflags)
    return suite



##############################################################################
#
# Copyright (c) 2010 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from setuptools import setup, find_packages

setup(
    name='z3c.checkversions',
    version='1.0',
    description="Find newer package versions on PyPI",
    long_description=(open("README.rst").read() + "\n" +
                      open("CHANGES.rst").read()),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Zope Public License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Quality Assurance',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Framework :: Buildout',
        'Framework :: Zope :: 2',
        'Framework :: Zope :: 3',
        'Framework :: Zope :: 4',
        'Framework :: Zope2',
        'Framework :: Zope3',
    ],
    keywords='version, buildout, packages, upgrade, zope, ztk',
    author='Zope Foundation and Contributors',
    author_email='zope-dev@zope.org',
    url='http://pypi.python.org/pypi/z3c.checkversions',
    license='ZPL 2.1',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    namespace_packages=['z3c'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'setuptools >= 8',
    ],
    extras_require={
        'buildout': ['zc.buildout'],
    },
    tests_require=['zc.buildout'],
    test_suite='z3c.checkversions.test.test_suite',
    entry_points="""
    [console_scripts]
    checkversions = z3c.checkversions.main:main
    """,
)



##############################################################################
#
# Copyright (c) 2010 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import pkg_resources
from z3c.checkversions import base

class Checker(base.Checker):
    """Checker class for installed packages
    """
    def get_versions(self, level=0):
        working_set = pkg_resources.working_set
        versions = dict([(d.key, d.version) for d in working_set])
        print("# Checking your installed distributions")
        return versions





##############################################################################
#
# Copyright (c) 2010 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""This script will check new package versions of either
your current installed distributions or a buildout file if provided.
It can detect major or minor versions availability:
level 0 gets the highest version (X.y.z),
level 1 gets the highest intermediate version (x.Y.z),
level 2 gets the highest minor version (x.y.Z).

Using level 2, you can automatically retrieve all bugfix versions of a buildout.

If you provide a blacklist file with bad versions, these versions won't be
suggested.
"""

from __future__ import absolute_import

from optparse import OptionParser
import os

def main():

    usage = u"Usage: %prog [-v] [-l LEVEL] [-i INDEX] [-b BLACKLIST] [-1] [buildout_file]"
    parser = OptionParser(description=__doc__, usage=usage)

    parser.add_option('-l', '--level',
                      type='int',
                      dest='level',
                      default=0,
                      help=u"Version level to check")

    parser.add_option('-i', '--index',
                      dest='index',
                      help=u"Provide and alternative package index URL")

    parser.add_option('-b', '--blacklist',
                      dest='blacklist',
                      default="",
                      help=u"Provide a blacklist file with bad versions")

    parser.add_option('-1', '--incremental',
                      dest='incremental',
                      action='store_true',
                      default=False,
                      help=u"Suggest only one upgrade. Skip others.")

    parser.add_option('-v', '--verbose',
                      dest='verbose',
                      action='store_true',
                      default=False,
                      help=u"Verbose mode (prints old versions too)")
    options, args = parser.parse_args()

    if len(args) > 1:
        parser.error("You must specify only one argument")

    if options.blacklist != "" and not os.path.exists(options.blacklist):
        parser.error('The blacklist file "%s" does not exist!' % options.blacklist)

    buildoutcfg = False
    if len(args) == 1:
        buildoutcfg = args[0]

    kw = {}
    if options.index is not None:
        kw['index_url'] = options.index

    if buildoutcfg:
        from . import buildout
        checker = buildout.Checker(filename=buildoutcfg,
                                   blacklist=options.blacklist,
                                   incremental=options.incremental,
                                   verbose=options.verbose,
                                   **kw)
    else:
        from . import installed
        checker = installed.Checker(blacklist=options.blacklist,
                                    incremental=options.incremental,
                                    verbose=options.verbose)

    checker.check(level=options.level)


if __name__ == '__main__':
    main()

